"""Invertible Checkpoint"""
import functools

import torch
from torch.utils.checkpoint import checkpoint as _checkpoint

from ..random import DelayedRNGFork
from ..utils.tools import pack, requires_grad

__all__ = ['checkpoint']


def get_tensor_id(inputs, by_storage=True):
    """Get a uniquely identifying key for a tensor based on its storage"""
    assert torch.is_tensor(inputs)
    return inputs.storage().data_ptr() if by_storage else id(inputs)


def get_tensor_id_set(*inputs, by_storage=True):
    """Get a set of only the tensors' ids"""
    get_id = functools.partial(get_tensor_id, by_storage=by_storage)
    return set(map(get_id, filter(torch.is_tensor, inputs)))


def checkpoint(
        function,
        *inputs,
        seed=True,
        strict=False,
        enabled=True,
        inverse=None,
        keep=(),
):
    """Same as `torch.utils.checkpoint.checkpoint` with extra functionalities

    The original checkpoint needs at least one input with `requires_grad` set
    to `True` to work. This is fine if `function` doesn't output tensors that
    require gradients due to some hidden parameters with `requires_grad=True`.
    In such case, a UserWarning will be raised and the code will run as if it
    was run in `torch.no_grad()` mode.

    Moreover, by default all output tensors will have `requires_grad=True` even
    if they shouldn't. This is because, in the forward pass, `function` will be
    called in `no_grad` mode and it is difficult to cheaply keep track of which
    output will require gradients beforehand.
    Refer to https://pytorch.org/docs/1.10.0/checkpoint.html for more details.

    Instead, this function has two new flags; `strict` and `enabled`.
    Setting `enabled` to `False`, will disable all checkpointing logic.
    However, when `strict=False`, it will use the original checkpoint as is.
    Else, when `strict=True`, the original behavior will change in two ways:
        (1) When no input tensor requires gradient but `function` generates
            outputs that do require gradients, the checkpoint will still work
        (2) All outputs will have `requires_grad` set to `False` by default.
            To specify what tensors actually require gradients, `function` will
            expect a keyword argument `strict` which will be `True` if we
            cannot automatically track this information. Here is an example:
    ```python
    def function(x, y, strict=None):
        z = x + y
        if strict:  # set requires_grad for all outputs
            z.requires_grad = x.requires_grad or y.requires_grad
            # no need to set for y as it is already set
        return z, y
    ```
           Debug your code carefully and try to cover all the cases. Don't
           forget to account for any used parameters that requires_grad.

    In addition, this function allows a more extreme version of checkpointing.
    If `function` was invertible with respect to its input tensors, then we can
    deallocate and recover them later using `inverse`. It should be a function
    that takes `function`'s outputs and returns computed `function`'s inputs.
    In general, `inverse` doesn't need to be differentiable and it'll run in
    `torch.inference_mode()`. It shouldn't have side effects and mustn't modify
    the outputs (its inputs) in-place in most cases. It only needs to compute
    the input tensors and can return anything for non-tensor inputs but it is
    a good practice to return them to allow for nesting invertible functions.

    For more granular control, it also expects an additional keyword argument
    `saved` which is the set of the positions of the inputs that are in memory.
    Here is the inverse of the previous example:
    ```python
    def inverse(z, y, saved=()):
        x = z - y if 0 in saved else None
        return x, y
    ```
    You might notice that `1` (for `y`) will always be in `saved` since it is
    an output of `function` and might be used in later operations. This was
    detected automatically by checking if any output share the same memory with
    an input. To manually select tensors to keep in memory, you can pass them
    to the `keep` argument as a tuple of tensors that are in the input already.

    There are few caveats to consider though. Invertible functions are hard to
    define without requiring more memory. Moreover, they are prone to numerical
    instabilities (e.g., multiplying by numbers close to zero). Even if we can
    get away with these fundamental problems, there are technical details to
    consider here. There is no way of guarding against accessing the data in
    the input tensors after calling the function and before the backward pass.
    It is up to the user to ensure this. Otherwise, it is possible to run into
    illegal memory access errors. Think of residual connections as an example.
    In `x + f(x)`, assuming `f` is an invertible checkpoint, `x` will be freed
    from memory before the sum is computed. On the other hand, we can maybe
    use `x.clone() + f(x)` (not `f(x) + x.clone()`!) but now we have a copy of
    `x` in memory. It is recommended to encapsulate this inside `f` itself or
    use the simple checkpoint instead. Other alternatives exists and you should
    study your case carefully before deciding to use this. Fore instance, check
    out `torch.autograd.graph.saved_tensors_hooks()` and `graph.save_on_cpu()`.
    Source: https://github.com/xmodar/invtorch

    Args:
        function: this is any forward function with positional arguments
        *inputs: input arguments tuple to be passed to `function`
        seed: same as preserve_rng_state; preserves the random number generator
        strict: `requires_grad` for outputs is set manually in `function`
        enabled: disables checkpointing if set to False
        inverse: inverse of `function` to deallocate the inputs as well
        keep: set of input tensors to keep in memory during inverse mode

    Returns:
        Outputs of `function(*inputs)` with checkpointing if required
    """
    if not enabled or not torch.is_grad_enabled():  # no checkpointing
        return function(*inputs)

    # find out which mode to run? nonstrict, strict, or strict inverse
    strict = strict or inverse is not None  # inverse is always strict

    if not strict:  # use torch.utils.checkpoint.checkpoint
        return _checkpoint(function, *inputs, preserve_rng_state=seed)

    return CheckpointFunction.apply(function, inverse, keep, seed, *inputs)


class CheckpointFunction(torch.autograd.Function):
    """Improved `torch.utils.checkpoint.CheckpointFunction` (invertible)"""
    # pylint: disable=abstract-method, arguments-differ, protected-access
    @classmethod
    def apply(cls, function, inverse, keep, seed, *inputs):
        """Refer to `invtorch.checkpoint()` for the arguments' details"""
        dif = torch.tensor((), requires_grad=True)  # ensures differentiability
        return super().apply(function, inverse, pack(keep), seed, dif, *inputs)

    @staticmethod
    def forward(ctx, function, inverse, keep, seed, _, *args):
        # capture autocast and RNG states and run the function
        ctx.set_materialize_grads(False)
        ctx.autocast = torch.is_autocast_enabled()
        devices = (x.device for x in args if torch.is_tensor(x) and x.is_cuda)
        ctx.forked_rng = DelayedRNGFork(devices, seed)
        outputs = function(*args, strict=True)
        assert isinstance(outputs, (torch.Tensor, tuple))

        # bookkeep differentiable tensors
        ctx.grads = list(map(requires_grad, pack(outputs)))
        no_grads = (x for x, g in zip(pack(outputs), ctx.grads) if not g)
        ctx.mark_non_differentiable(*filter(torch.is_tensor, no_grads))
        if not any(ctx.grads):
            return outputs  # apparently, `function` was not differentiable

        # get all tensors that should be deallocated from memory
        if inverse is None:
            seep = ()
        else:
            keep = get_tensor_id_set(*keep, *pack(outputs))
            seep = get_tensor_id_set(*args) - keep
            inverse = inverse if seep else None
        ctx.function, ctx.inverse = function, inverse

        # separate kept and deallocated tensors from other inputs
        ctx.indices, tensors, ctx.deallocated, ctx.inputs = [], [], set(), {}
        for i, argument in enumerate(args):
            if torch.is_tensor(argument):
                ctx.indices.append(i)
                tensors.append(argument)
                if get_tensor_id(argument) in seep:
                    argument.storage().resize_(0)  # deallocate the tensor
                    ctx.deallocated.add(i)
            else:
                ctx.inputs[i] = argument
        ctx.save_for_backward(*tensors)
        if seep:
            ctx.outputs = outputs
        return outputs

    @staticmethod
    def backward(ctx, *args):
        if not torch.autograd._is_checkpoint_valid():
            raise RuntimeError('.grad()/.backward(inputs=...) not supported')

        # materialize any deallocated tensors by calling inverse
        tensors = ctx.saved_tensors
        if ctx.inverse is not None:
            saved = set(ctx.indices) - ctx.deallocated
            kwargs = dict(saved=saved) if saved else {}
            with torch.inference_mode():
                inverted = pack(ctx.inverse(*pack(ctx.outputs), **kwargs))
                for i, idx in enumerate(ctx.indices):
                    if idx in ctx.deallocated:
                        tensors[i].set_(inverted[idx])
            ctx.inverse = ctx.outputs = inverted = None

        # detach input tensors and run function again but in grad_mode
        inputs = ctx.inputs.copy()
        for i, idx in enumerate(ctx.indices):
            inputs[idx] = tensors[i].detach()
            inputs[idx].requires_grad_(tensors[i].requires_grad)
        inputs = [inputs[i] for i in range(len(inputs))]
        with torch.enable_grad(), torch.cuda.amp.autocast(ctx.autocast):
            with ctx.forked_rng:
                outputs = pack(ctx.function(*inputs, strict=False))

        # check if requires_grad matches that of function call in forward
        current = map(requires_grad, outputs)
        bad = {i for i, (g, c) in enumerate(zip(ctx.grads, current)) if g != c}
        if bad:
            expected = [(i in bad) != g for i, g in enumerate(ctx.grads)]
            msg = ('manually set requires_grad for output tensors in '
                   'strict mode mismatched automatically set values in '
                   'the backward pass. Please, debug your implementation '
                   'carfully and try to cover all the cases. Keep in mind '
                   'the paramters of any model you call in `function`.'
                   f'\nExpected: {expected}\nReceived: {ctx.grads}')
            raise RuntimeError(msg)

        # perform the backward pass on outputs that requires_grad
        outputs_with_grad, args_with_grad = [], []
        for output, arg in zip(outputs, args):
            if arg is not None and requires_grad(output):
                outputs_with_grad.append(output)
                args_with_grad.append(arg)
        torch.autograd.backward(outputs_with_grad, args_with_grad)
        grads = (x.grad if torch.is_tensor(x) else None for x in inputs)
        return (None, None, None, None, None, *grads)



"""InvTorch: Memory-Efficient Invertible Functions

Sources:
    https://github.com/xmodar/invtorch
    https://github.com/silvandeleemput/memcnn
    https://gist.github.com/xmodar/4deb8905ed8c294862972466f69e5d17
    https://gist.github.com/xmodar/2328b13bdb11c6309ba449195a6b551a
    https://gist.github.com/xmodar/7921460648230eda5053fe06b7cd2f4d
"""
from . import nn, random, utils
from .utils.checkpoint import checkpoint

__version__ = '0.3.1'



"""Base Invertible Modules"""
import itertools
from contextlib import contextmanager

import torch
from torch import nn

from ...utils.checkpoint import checkpoint
from ...utils.tools import pack, requires_grad

__all__ = ['Module', 'WrapperModule']


class Module(nn.Module):
    """Base invertible module `inputs = self.inverse(*self.function(*inputs))`

    Use this with great caution. Refer to the notes in `invtorch.checkpoint()`
    Source: https://github.com/xmodar/invtorch
    """
    def __init__(self):
        super().__init__()
        self.seed = False  # preserve RNG state in backward
        self.checkpoint = True  # enables or disables checkpointing
        self.invertible = True  # use inverse if checkpointing is enabled
        self._reversed = False  # switch function and inverse

    def function(self, *inputs, strict=None, saved=()):
        """Compute the outputs of the function given the inputs"""
        raise NotImplementedError

    @property
    def call_function(self):
        """Current function (according to `self.reversed`)"""
        return self.inverse if self.reversed else self.function

    def inverse(self, *outputs, strict=None, saved=()):
        """Compute the inputs of the function given the outputs"""
        raise NotImplementedError

    @property
    def call_inverse(self):
        """Current inverse (according to `self.reversed`)"""
        return self.function if self.reversed else self.inverse

    num_function_outputs = num_inverse_outputs = None

    @property
    def num_outputs(self):
        """Current number of outputs (according to `self.reversed`)"""
        if self.reversed:
            return self.num_inverse_outputs
        return self.num_function_outputs

    @property
    def reversible(self):
        """Whether function and inverse can be switched"""
        return False

    def reverse(self, mode=None):
        """Switch function and inverse"""
        if not self.reversed if mode is None else mode:
            assert self.reversible, 'module is not reversible'
            self._reversed = True
        else:
            self._reversed = False
        return self

    def forward(self, *inputs, **kwargs):
        """Perform the forward pass"""
        private = {
            'seed': self.seed,
            'enabled': self.checkpoint,
            'inverse': self.call_inverse if self.invertible else None,
        }
        assert all(k not in kwargs for k in private), 'got an illegal argument'
        kwargs.setdefault('strict', True)
        kwargs.update(private)
        outputs = checkpoint(self.call_function, *inputs, **kwargs)
        return self.process_outputs(*pack(outputs))

    def process_outputs(self, *outputs):
        """Get only `self.forward()` outputs"""
        num_outputs = self.num_outputs
        if num_outputs is None:
            num_outputs = len(outputs)
        elif num_outputs < 0:
            num_outputs += len(outputs)
        assert 0 < num_outputs <= len(outputs), f'need {num_outputs} outputs'
        assert not requires_grad(any=outputs[num_outputs:]), (
            'discarded outputs must not be differentiable')
        return outputs[0] if num_outputs == 1 else outputs[:num_outputs]

    @property
    def checkpoint(self):
        """Whether the module is in checkpoint or pass_through mode"""
        return self._checkpoint

    @checkpoint.setter
    def checkpoint(self, value):
        if value:
            self._checkpoint = True
        else:
            self._checkpoint = self._invertible = False

    @property
    def invertible(self):
        """Whether the module is in invertible or simple checkpoint mode"""
        return self._checkpoint and self._invertible

    @invertible.setter
    def invertible(self, value):
        if value:
            self._invertible = self._checkpoint = True
        else:
            self._invertible = False

    @property
    def reversed(self):
        """Whether function and inverse should be switched"""
        return self._reversed

    @reversed.setter
    def reversed(self, value):
        self.reverse(value)

    def check_function(self, *inputs):
        """Check if `self.call_function()` is consistent when `strict=True`"""
        with torch.enable_grad():
            outputs1 = pack(self.call_function(*inputs, strict=False))
            outputs1 = pack(self.process_outputs(*outputs1))
        with torch.no_grad():
            outputs2 = pack(self.call_function(*inputs, strict=True))
            outputs2 = pack(self.process_outputs(*outputs2))
        assert len(outputs2) == len(outputs1), 'number of outputs'
        grads1 = list(map(requires_grad, outputs1))
        grads2 = list(map(requires_grad, outputs2))
        bad = {i for i, (g1, g2) in enumerate(zip(grads1, grads2)) if g1 != g2}
        expected = [(i in bad) != g for i, g in enumerate(grads2)]
        assert not bad, f'Received: {grads2}\nExpected: {expected}'
        return True

    @torch.no_grad()
    def check_inverse(self, *inputs, atol=1e-5, rtol=1e-3):
        """Check if `self.call_inverse()` computes correct input tensors"""
        outputs = pack(self.call_function(*inputs, strict=True))
        outputs = pack(self.call_inverse(*outputs))
        for inputs, outputs in itertools.zip_longest(inputs, outputs):
            is_tensor = torch.is_tensor(inputs)
            assert is_tensor == torch.is_tensor(outputs), 'out types mismatch'
            same = not is_tensor or torch.allclose(inputs, outputs, rtol, atol)
            assert same, 'an inverted tensor mismatched (try double precision)'
        return True

    def get_extra_state(self):
        return {
            'seed': self.seed,
            'checkpoint': self.checkpoint,
            'invertible': self.invertible,
            'reversed': self.reversed,
        }

    def set_extra_state(self, state):
        self.seed = state['seed']
        self.checkpoint = state['checkpoint']
        self.invertible = state['invertible']
        self.reversed = state['reversed']

    @contextmanager
    def temp_mode(self, **kwargs):
        """Set, temporarily, the mode of the model"""
        state = {}
        for key in ('seed', 'checkpoint', 'invertible', 'reversed'):
            state[key] = getattr(self, key)
            if key in kwargs and state[key] == bool(kwargs[key]):
                kwargs.pop(key)
        assert all(k in state for k in kwargs), 'got an illegal argument'
        if 'checkpoint' in kwargs and 'invertible' in kwargs:
            assert kwargs['checkpoint'] or not kwargs['invertible'], (
                'set either `checkpoint` or `invertible` or avoid conflict')
        try:
            for key, value in kwargs.items():
                setattr(self, key, value)
            yield self
        finally:
            for key, value in state.items():
                setattr(self, key, value)

    def extra_repr(self):
        extra = f'reversed={self.reversed}, checkpoint={self.checkpoint}'
        if self.checkpoint:
            extra += f', invertible={self.invertible}, seed={self.seed}'
        return extra


class WrapperModule(Module):
    """Base wrapper invertible module"""
    # pylint: disable=abstract-method
    wrapped_type = ()

    def __init__(self, module):
        assert self.wrapped_type, 'define wrapped_type'
        assert isinstance(module, self.wrapped_type), (
            f'{type(module).__name__} is not in <{self.wrapped_type}>')
        super().__init__()
        self.module = module

    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except AttributeError:
            return getattr(self.module, name)



"""Invertible Tools"""
import builtins

import torch

__all__ = ['pack', 'requires_grad']


def pack(inputs, is_tensor=False):
    """Pack the inputs into tuple if they were a single tensor"""
    single = torch.is_tensor(inputs)
    outputs = (inputs, ) if single else inputs
    return (outputs, single) if is_tensor else outputs


def requires_grad(inputs=None, *, any=None, all=None):
    """Check if the inputs is a tensor that requires_grad"""
    # pylint: disable=redefined-builtin
    if any is not None or all is not None:
        any = any is None or builtins.any(map(requires_grad, any))
        all = all is None or builtins.all(map(requires_grad, all))
        requires = any and all
        if inputs is not None:
            inputs.requires_grad_(requires)
        return requires
    assert inputs is not None, 'You have to pass inputs'
    return torch.is_tensor(inputs) and inputs.requires_grad



"""Invertible Flatten Modules"""
from .module import Module

__all__ = ['Flatten']


class Flatten(Module):
    """Invertible flatten operation"""
    def __init__(self, start_dim=1, end_dim=-1, unflattened_size=None):
        super().__init__()
        self.start_dim, self.end_dim = start_dim, end_dim
        self.unflattened_size = unflattened_size

    def function(self, inputs, *, strict=None):
        # pylint: disable=arguments-differ, unused-argument
        end_dim = self.end_dim
        if end_dim < 0:
            end_dim += inputs.dim()
        if self.end_dim == inputs.dim():
            end_dim = None
        else:
            end_dim += 1
        unflattened_size = inputs.shape[self.start_dim:end_dim]
        if self.unflattened_size is not None:
            assert tuple(self.unflattened_size) == unflattened_size, 'mismatch'
        outputs = inputs.flatten(self.start_dim, self.end_dim)
        if strict:
            outputs.requires_grad_(inputs.requires_grad)
        return outputs, unflattened_size

    def inverse(self, outputs, unflattened_size=None, *, strict=None):
        # pylint: disable=arguments-differ, unused-argument
        if unflattened_size is None:
            unflattened_size = self.unflattened_size
        inputs = outputs.unflatten(self.start_dim, unflattened_size)
        if strict:
            inputs.requires_grad_(outputs.requires_grad)
        return inputs

    num_function_outputs = 1

    @property
    def reversible(self):
        return self.unflattened_size is not None



"""Invertible Container Modules"""
from torch import nn

from ...utils.tools import pack
from .module import WrapperModule

__all__ = ['Sequential']


class Sequential(WrapperModule):
    """Sequential Invertible module"""
    wrapped_type = nn.Sequential

    def __init__(self, *modules):
        super().__init__(nn.Sequential(*modules))

    def function(self, *inputs, strict=None):
        # pylint: disable=arguments-differ
        extras, counts = [], []
        for layer in self.module:
            outputs = pack(layer.call_function(*inputs, strict=strict))
            inputs = pack(layer.process_outputs(*outputs))
            counts.append(len(outputs) - len(inputs) if strict else 0)
            if strict:
                extras.extend(outputs[len(inputs):])
        return (*inputs, *extras, counts)

    def inverse(self, *outputs, saved=()):
        # pylint: disable=arguments-differ
        extras, end = [()] * len(outputs[-1]), -1
        for i, count in enumerate(reversed(outputs[end]), 1):
            extras[-i] = outputs[end - count:end]
            end -= count
        outputs = outputs[:end]
        kwargs = {}
        for layer in reversed(self.module):
            outputs = outputs + extras.pop()
            if not extras and saved:
                kwargs['saved'] = saved
            outputs = pack(layer.call_inverse(*outputs, **kwargs))
        return outputs[0] if len(outputs) == 1 else outputs

    def process_outputs(self, *outputs):
        return super().process_outputs(*outputs[:-sum(outputs[-1]) - 1])

    call_function, call_inverse = function, inverse

    @property
    def reversible(self):
        return all(layer.reversible for layer in self.module)

    def reverse(self, mode=None):
        if self.reversed != super().reverse(mode).reversed:
            layers = reversed(layer.reverse() for layer in self.module)
            self.module = nn.Sequential(*layers)



"""Random Number Generator Utilities"""
import torch

__all__ = ['DelayedRNGFork']


class DelayedRNGFork:
    """Caputres PyTorch's RNG state on initialization with delayed use

    This context manager captures the torch random number generator states on
    instantiation. Then, it can be used many times later using with-statements

    Example:
        rng = DelayedRNGFork(devices=[0])
        print('1 outside', torch.randn(3, device='cuda:0'))
        with rng:
            print('1 inside ', torch.randn(3, device='cuda:0'))
            print('2 inside ', torch.randn(3, device='cuda:0'))
        print('2 outside', torch.randn(3, device='cuda:0'))
    """
    def __init__(self, devices=None, enabled=True):
        self.cpu_state = torch.get_rng_state() if enabled else None
        self.gpu_states = {}
        if self.enabled:
            if devices is None:
                devices = range(torch.cuda.device_count())
            for device in map(torch.cuda.device, set(devices)):
                self.gpu_states[device.idx] = torch.cuda.get_rng_state(device)
        self._fork = None

    def __enter__(self):
        if self.enabled:
            self._fork = torch.random.fork_rng(self.gpu_states)
            self._fork.__enter__()  # pylint: disable=no-member
            torch.set_rng_state(self.cpu_state)
            for device, state in self.gpu_states.items():
                torch.cuda.set_rng_state(state, device)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.enabled:
            fork, self._fork = self._fork, None
            # pylint: disable=no-member
            return fork.__exit__(exc_type, exc_value, traceback)
        return None

    @property
    def enabled(self):
        """Whether the fork is enabled"""
        return self.cpu_state is not None



"""Linear Invertible Modules"""
import functools

from torch import nn
from torch.nn import functional as F

from ...utils.tools import requires_grad
from .module import Module, WrapperModule

__all__ = ['Identity', 'Linear']


class Identity(Module):
    """Identity function"""
    reversible = True

    def forward(self, *inputs, **kwargs):
        return inputs[0] if len(inputs) == 1 else inputs

    function = inverse = call_function = call_inverse = forward


class Linear(WrapperModule):
    """Invertible linear module"""
    wrapped_type = nn.Linear

    @functools.wraps(nn.Linear.__init__)
    def __init__(self, *args, **kwargs):
        super().__init__(nn.Linear(*args, **kwargs))
        assert self.in_features <= self.out_features, 'few out_features'

    @property
    def reversible(self):
        return self.in_features == self.out_features

    def function(self, inputs, *, strict=None):
        """Compute the outputs of the function"""
        # pylint: disable=arguments-differ
        outputs = F.linear(inputs, self.weight, self.bias)
        if strict:
            requires_grad(outputs, any=(inputs, self.weight, self.bias))
        return outputs

    def inverse(self, outputs, *, strict=None):
        """Compute the inputs of the function"""
        # pylint: disable=arguments-differ
        if self.bias is not None:
            outputs = outputs - self.bias
        inputs = F.linear(outputs, self.weight.pinverse())
        if strict:
            requires_grad(inputs, any=(outputs, self.weight, self.bias))
        return inputs



"""Invertible Convolutions"""
import functools

import torch
import torch.nn.functional as F
from torch import nn

from ...utils.tools import requires_grad
from .module import WrapperModule

__all__ = ['Conv1d', 'Conv2d', 'Conv3d']


class _ConvNd(WrapperModule):
    """Invertible convolution"""
    wrapped_type = nn.modules.conv._ConvNd  # pylint: disable=protected-access

    def __init__(self, module):
        super().__init__(module)
        outputs, inputs = self.flat_weight_shape
        assert inputs <= outputs, f'out_channels/groups={outputs} < {inputs}'
        # TODO: assert kernel_size is still invertible given stride & dilation

    @property
    def reversible(self):
        outputs, inputs = self.flat_weight_shape
        return inputs == outputs, f'out_channels/groups={outputs} != {inputs}'

    def function(self, inputs, *, strict=None):
        # pylint: disable=arguments-differ
        # TODO: make input padding an opt-in feature
        input_padding = self.get_input_padding(inputs.shape)
        assert sum(input_padding) == 0, f'inputs need padding: {inputs.shape}'
        outputs = self.module.forward(inputs)
        if strict:
            requires_grad(outputs, any=(inputs, self.weight, self.bias))
        return outputs

    def inverse(self, outputs, *, strict=None):
        # pylint: disable=arguments-differ
        old_outputs = outputs
        if self.bias is not None:
            outputs = outputs - self.bias.view(-1, *[1] * self.dim)

        # compute the overlapped inputs
        factor, input_shape = self.flat_weight_shape
        weight = self.weight.view(self.groups, -1, input_shape)
        weight = weight.pinverse().transpose(-1, -2).reshape_as(self.weight)
        inputs = self.conv_transpose(outputs, weight)

        # TODO: make kernel overlaps an opt-in feature
        # normalize the overlapping regions
        one = torch.ones((), device=inputs.device, dtype=inputs.dtype)
        outputs = (one / factor).expand(1, *outputs.shape[1:])
        overlaps_weight = one.expand(self.out_channels, 1, *self.kernel_size)
        overlaps = self.conv_transpose(outputs, overlaps_weight)
        inputs = inputs.div_(overlaps)

        if strict:
            requires_grad(inputs, any=(old_outputs, self.weight, self.bias))
        return inputs

    def get_input_padding(self, input_size):
        """Get the input padding given the input size"""
        def rule(size, kernel_size, stride, padding, dilation):
            margin = 2 * padding - dilation * (kernel_size - 1) - 1
            return size - ((size + margin) // stride) * stride + margin

        input_size = tuple(input_size)[-self.dim:]
        args = self.kernel_size, self.stride, self.padding, self.dilation
        return tuple(map(lambda x: rule(*x), zip(input_size, *args)))

    def conv_transpose(self, inputs, weight):
        """Compute conv_transpose"""
        args = None, self.stride, self.padding, 0, self.groups, self.dilation
        return getattr(F, f'conv_transpose{self.dim}d')(inputs, weight, *args)

    @property
    def dim(self):
        """Number of dimensions"""
        return self.weight.dim() - 2

    @property
    def flat_weight_shape(self):
        """Output and input shapes"""
        shape = self.weight.shape
        return shape[0] // self.groups, shape[1:].numel()


class Conv1d(_ConvNd):
    """Invertible 1D convolution"""
    wrapped_type = nn.Conv1d

    @functools.wraps(nn.Conv1d.__init__)
    def __init__(self, *args, **kwargs):
        super().__init__(nn.Conv1d(*args, **kwargs))


class Conv2d(_ConvNd):
    """Invertible 2D convolution"""
    wrapped_type = nn.Conv2d

    @functools.wraps(nn.Conv2d.__init__)
    def __init__(self, *args, **kwargs):
        super().__init__(nn.Conv2d(*args, **kwargs))


class Conv3d(_ConvNd):
    """Invertible 3D convolution"""
    wrapped_type = nn.Conv3d

    @functools.wraps(nn.Conv3d.__init__)
    def __init__(self, *args, **kwargs):
        super().__init__(nn.Conv3d(*args, **kwargs))
